-- | Privide declassification combinators (trusted).
module Declassification
( 
    Hatch
  , Open
  , Close
  , Authority 
  , hatch
  , ntimes
  , flock 
  , dlm
  , certify 
)

where

import Sec ( open, reveal )
import SecIO 
import Data.IORef
import Lattice ( Less )
import Sec ( Sec )


-- | Type for escape hatches.
type Hatch s s' a b = Sec s a -> SecIO s' b

-- | Used by 'flock'. 'Open' represents computations that open flow locks.
type Open  = IO ()
-- | Used by 'flock'. 'Close' represents computations that close flow locks.
type Close = IO ()

-- | Used by 'dlm'. 
data Authority s = Authority Open Close 


-- | Creates an escape hatch. 
hatch :: Less s' s => (a -> b) -> Hatch s s' a b  
hatch f = \sa -> return ( ( f (reveal sa) ) ) 

{- | Limite the number of times that an escape hatch can be applied by 
     a single run of the program.
-}
ntimes :: Int -> Hatch s s' a b -> IO (Hatch s s' a b)  
ntimes n f = do ref <- newIORef n
                return $ \sa ->
                             MkSecIO $
                              do k <- readIORef ref
                                 let MkSecIO m = f sa 
                                 if k <= 0
                                    then error "ntimes fails" 
                                    else do writeIORef ref (k-1)
                                            m  


{- | This function associates a flow lock to an escape hatch. 
     Then, the escape hatch 
     can be successfully applied when the flow lock is open. In contrast, 
     the escape hatch cannot by applied after closing the flock lock.
-}
flock :: Hatch s s' a b -> IO (Hatch s s' a b, Open, Close)  
flock f = do ref <- newIORef False 
             return (\sa -> MkSecIO $ 
                             do b <- readIORef ref
                                let MkSecIO m = f sa 
                                if b then m  
                                     else error "flock fails!"  
                     , writeIORef ref True 
                     , writeIORef ref False
                    )
 

{- | This function allows to an escape hatch to be applied only when 
     the running code can be certified with some authority. 
-}
dlm :: Hatch s s' a b -> IO (Hatch s s' a b, Authority s )  
dlm f = do (whof, open, close) <- flock f
           return (whof, Authority open close) 

-- | Certifies that a piece of code have certain authority.
certify :: s -> Authority s -> IO a -> IO a 
certify s (Authority open close) io = 
  s `seq` ( do open ; a <- io ; close ; return a )
