-- | Provide security for computations with side-effects. 
module SecIO where

import Lattice
import Sec

import Control.Monad
import System.Directory
import Data.IORef
import Network.Socket

-- Monad

{-| Secure side-effect computations. 
    These computations perform side-effects at security level s (or above) and return 
    a value of type a.
-}
newtype SecIO s a = MkSecIO (IO (Sec s a))

instance Functor (SecIO s) where
  h `fmap` (MkSecIO io) = MkSecIO ( do sec <- io
                                       return (h `fmap` sec) )


instance Monad (SecIO s) where
  return x =
    MkSecIO (return (return x))
  
  MkSecIO m >>= k =
    MkSecIO (do sa <- m
                let MkSecIO m' = k (reveal sa)
                m')

-- SecIO functions

-- | Lift a pure confidential value into a secure side-effect computation. 
value :: Sec s a -> SecIO s a
value sa = MkSecIO (return sa)


-- | Execution of secure computations.
run :: SecIO s a -> IO (Sec s a)
run (MkSecIO m) = m

-- | Safetly downgrade the security level that restrict the side-effects performed by the computations. 
plug :: Less sl sh => SecIO sh a -> SecIO sl (Sec sh a)
plug secio_sh@(MkSecIO m) = less sl sh `seq` secio_sl
                            where 
                                 (secio_sl) = MkSecIO (do sha <- m
                                                          return (sec sha))
                                 sl = unSecIOType secio_sl
                                 sh = unSecIOType secio_sh

-- | Internal function, not exported. For type-checking purposes.
unSecIOType :: SecIO s a -> s   
unSecIOType _ = undefined



{-| 
    Represent secure locations.
    Data type that is internally used to easily introduce new side-effects into SecIO.
    Type t is the raw type needed to perform side-effects. For instance, 
    t is 'FilePath' for file, 'IORef' a for writing and reading
    references, etc. Type s is the confidentiality level of the 
    location. Type a is the kind of values written and read form t. 
-}

type DataInvariant a = (a -> IO Bool)
data Loc t s a b = MkLoc t (DataInvariant a) (DataInvariant a)

invariant_input :: DataInvariant a -> Loc t s a b -> Loc t s a b
invariant_input di (MkLoc t _ o) = MkLoc t di o

invariant_output :: DataInvariant a -> Loc t s a b -> Loc t s a b
invariant_output di (MkLoc t i _) = MkLoc t i di


{-| Introduce unsecure read and write operations for side-effects of type t 
    that stores and read values of type a. 
-}
class UnsecureRW t a b | t -> a b where
   unsec_write :: t -> a -> IO b 
   unsec_read  :: t -> IO a       

{-| Introduce safe read and write operations for locations of type 'Loc' t s a.
    This class acts as a wrapper for the type class 'UnsecureRW'.
-}
class UnsecureRW t a b => SecureRW t s a b where
   {-| Secure read operation when the fact of just reading might be visible 
       for the attacker from inside of the program. 
       For example, reading from a channel in a network communication changes the 
       buffer pointer for that channel, which can be exploited to leak information. 
   -}
   effectful_read  :: Loc t s a b -> SecIO s a
   {-| Secure read operation when the fact of just reading is not visible by the attacker from 
       inside of the program.
       For example, reading from references does not produce any observable effect.
   -}
   effectless_read :: Loc t s a b -> SecIO s' (Sec s a)
   -- | Secure write operation.
   effectful_write :: Loc t s a b -> a -> SecIO s b

   -- Create read and write resource
   create_resource :: t -> Loc t s a b


instance UnsecureRW t a b => SecureRW t s a b where
    effectless_read (MkLoc loc i _)   = MkSecIO $ (sec.sec) `fmap` (do a <- unsec_read loc
    								       b <- i a
								       if b then return a else diError)
    	where diError = error "Data invariant violated while reading."
    effectful_read  (MkLoc loc i _)   = MkSecIO $ (sec      `fmap` (do a <- unsec_read loc
    								       b <- i a
								       if b then return a else diError))
    	where diError = error "Data invariant violated while reading."
    effectful_write (MkLoc loc _ o) a = MkSecIO $ (sec      `fmap` (do b <- o a
    								       if b then unsec_write loc a else diError))
	where diError = error "Data invariant violated while writing."
    create_resource t 		      = MkLoc t (\s -> return True) (\s -> return True)
	 
    
-- | Locations that represent files.  
type File s = Loc FilePath s String ()

instance UnsecureRW FilePath String () where
   unsec_read file      = readFile file
   unsec_write file str = writeFile file str


-- | Secure reading from a file.
readFileSecIO  :: File s -> SecIO s' (Sec s String)
readFileSecIO  = effectless_read

-- | Secure writing to a file.
writeFileSecIO :: File s -> String -> SecIO s ()
writeFileSecIO = effectful_write

makeFile :: FilePath -> File s
makeFile = create_resource

-- Creation of low files
newFileSecIO :: FilePath -> SecIO L (File L)
newFileSecIO f = MkSecIO $
     do b <- doesFileExist f
     	case b of
		True  -> error "File alredy exists."
		False -> return $ sec (create_resource f)

-- | Locations that represent references. 
type Ref s a = Loc (IORef a) s a ()

instance UnsecureRW (IORef a) a () where
   unsec_read  ref    = readIORef ref
   unsec_write ref a  = writeIORef ref a


-- | Secure reading from a reference.
readRefSecIO  :: Ref s a -> SecIO s' (Sec s a)
readRefSecIO  = effectless_read 

-- | Secure writing to a reference.
writeRefSecIO :: Ref s a -> a -> SecIO s ()
writeRefSecIO = effectful_write


{-| Secure creation of a reference. We assume that the attacker 
    has no manner to observe the side-effect of creating a reference from inside of the program, 
    for example, by inspecting the free memory. At the moment, there are no such functions inside of 'SecIO'.
    Nevertheless, if there is a consideration of include, for instance, a function that returns 
    the free memory in the program, then the function type of 'newIORefSecIO' needs 
    to be changed. 
-}
newIORefSecIO :: a -> SecIO s' (Ref s a) 
newIORefSecIO a = MkSecIO ( (sec.create_resource) `fmap` newIORef a )


{-|  Location that represents the screen-keywoard. Here, we can choose between 
     a) The attacker is on the screen-keyword, which implies that when taking input 
     from the screen has an effect -- the attacker can detect the input. Therefore,
     we need to implement taking input from the keyword using 'effectful_read'.
     b) The attacker is not on the screen-keyword. In this case, we implement 
     reading using 'effectless_read'. We choose option a) as a model.
-}
type Screen = Loc () L String ()


instance UnsecureRW () String () where
   unsec_read  _    = getLine 
   unsec_write _ a  = putStr a


-- | Secure input from the keyword.
getLineSecIO  :: Screen -> SecIO L String
getLineSecIO  = effectful_read 

-- | Secure output to the screen.
putStrSecIO :: Screen -> String -> SecIO L ()
putStrSecIO = effectful_write


-- | Derived operations for reading from files (legacy code).
s_read :: Less s' s => File s' -> SecIO s String
s_read file =
   do ss <- readFileSecIO file
      value (up ss)

-- | Derived operations for writing into files (legacy code).
s_write :: Less s s' => File s' -> String -> SecIO s (Sec s' ())
s_write file s = plug (writeFileSecIO file s)

makeScreen :: Screen
makeScreen = create_resource ()


{------------------------------------
 --- Network operations
-------------------------------------}

type SecSocket s   = Loc (Socket,Int) s String Int  


{- Data types needed to create a connection -}
{- Which connection is high or low is defined in the 
   trusted code!!!
-}
data SecSockAddr s = MkSockAddr SockAddr


inet_addrSecIO :: String -> SecIO s HostAddress
inet_addrSecIO s = MkSecIO $ (return `fmap` (inet_addr s))


socketSecIO :: Family -> SocketType -> ProtocolNumber -> 
               SecIO s' (SecSocket s)
socketSecIO f s p =  MkSecIO ( (sec.(\s -> create_resource (s,0))) `fmap` socket f s p)


-- Question: it might be that it is possible to return 
-- SecIO s' () if it is the case that sIsBoundSecIO is the 
-- only observer if a socket is bind or not.
bindSocketSecIO :: SecSocket s -> SecSockAddr s -> SecIO s ()
bindSocketSecIO (MkLoc (s,_) i o) (MkSockAddr address) = 
                 MkSecIO (sec `fmap` bindSocket s address) 


sIsBoundSecIO :: SecSocket s -> SecIO s' (Sec s Bool)
sIsBoundSecIO (MkLoc (s,_) i o) = MkSecIO (sec `fmap` (sec `fmap` sIsBound s)) 


acceptSecIO :: SecSocket s -> SecIO s (SecSocket s, SecSockAddr s)
acceptSecIO (MkLoc (s,_) i o) = MkSecIO (sec `fmap` m)
                            where m = do (sock, addr) <- accept s   
                                         return (MkLoc (sock,0) i o, MkSockAddr addr)


instance UnsecureRW (Socket,Int) String Int where
  unsec_read  (sock,n)     = recv sock n  
  unsec_write (sock,_) str = send sock str   


recvSecIO :: SecSocket s -> Int -> SecIO s String
recvSecIO (MkLoc (s,_) i o) n   = effectful_read (MkLoc (s,n) i o)


sendSecIO :: SecSocket s -> String -> SecIO s Int
sendSecIO = effectful_write 
